import java_bytecode;

(* Definitions *)
set locals = seq(local,$max_locals);
set local_vars = seq(var(l),$max_locals);
set field_vars = seq(var(f),$nb_fields);
set heap_vars = seq(var(h),$nb_classes);
set pcs = seq(pp,$max_pc);

(* Creation of objects constructors of type object0(var(f0),...), ... *)
(* and object0(nilfield,...), ... *)
(* and objectO(zero,...), object1(val_null,...), ... representing objects *)
(* initialized by Java.lang.Object constructor *)
set objects_vars = cseq(val_null,$nb_classes);
set initialized_objects = cseq(val_null,$nb_classes);

let f_used = 0 in (
  for i from 1 to $nb_classes
  do(
    let nb_fields_i = getn($nb_fields_per_classes,$i) in (
      let field_vars_i = seq(var(f),$nb_fields_i)
      and init_fields_i = getn($init_fields_per_classes,$i) in (
	for j from 1 to $nb_fields_i
	do(
(*********************************************)
	  set f_used = add(1,$f_used);
	  let f_var = getn($field_vars,$f_used) in (
	    setn(field_vars_i,$j) = $f_var;
	  );
	);
	let o = add($i,-1) in (
	  setn(objects_vars,$i) = concat(object,($o,$field_vars_i));
	  setn(initialized_objects,$i) = concat(object,($o,$init_fields_i));
(********************************************)
	);
      );
    );
  );
);

defun get_object_var(i) =
(
  getn($objects_vars,$i);
);

defun get_initialized_object(i) =
(
  getn($initialized_objects,$i);
);

(* Creation of nilstacks constructors of type nilstack0 ... *)
set nilstacks = seq(nilstack,$nb_classes);
defun get_nilstack(i) =
(
  getn($nilstacks,$i);
);

(* Creation of nillocal sequence *)
set nillocals = locals(cseq(nillocal,$max_locals));

(* Creation of object stacks constructors without parameters of type *)
(* stack0(ni, ... *)
set stacks = useq(stack,$nb_classes);
defun get_stack(i) =
(
  getn($stacks,$i);
);

(* Creation of nillock *)

(***********************************************)
set nillock = lock(0,0);
(***********************************************)

(* alias for main class *)
set main_class = getn($classes,1);

(* alias for initialization method *)
set init_method = Method(<<init>>,NilType,void);

(* alias for stubs *)
set java_lang_string = Class(ConsName(<java>,ConsName(<lang>,ConsName(<String>,NilName))));
set java_lang_stringbuilder = Class(ConsName(<java>,ConsName(<lang>,ConsName(<StringBuilder>,NilName))));
set java_lang_object = Class(ConsName(<java>,ConsName(<lang>,ConsName(<Object>,NilName))));
set java_lang_system = Class(ConsName(<java>,ConsName(<lang>,ConsName(<System>,NilName))));
set java_io_exception = Class(ConsName(<java>,ConsName(<io>,ConsName(<Exception>,NilName))));
set java_io_inputstream = Class(ConsName(<java>,ConsName(<io>,ConsName(<InputStream>,NilName))));
set java_io_printstream = Class(ConsName(<java>,ConsName(<io>,ConsName(<PrintStream>,NilName))));
set java_lang_thread = Class(ConsName(<java>,ConsName(<lang>,ConsName(<Thread>,NilName))));

(* definition of initialJavaState *)
set iframe = frame(name(Method(<main>,ConsType(AType(
OType($java_lang_string)),NilType),void),$main_class),pp0,nilstack,$nillocals);


(*************************************)
set iheap_args = $nilstacks;
for i from 1 to $nb_classes
do(
  setn(iheap_args,$i) = heap(0,getn($nilstacks,$i));
);
(******************************************)

let input_class_num = mem($java_io_inputstream,$classes)
and print_class_num = mem($java_io_printstream,$classes) in (
  let input_obj = $get_initialized_object($input_class_num)
  and print_obj = $get_initialized_object($print_class_num) in (
(****************************************************)
    let input_num = add($input_class_num,-1)
    and print_num = add($print_class_num,-1) in (
(****************************************************)
      let input_stack = concat(stack,($input_num,($input_obj,$nillock,getn($nilstacks,$input_class_num))))
      and print_stack = concat(stack,($print_num,($print_obj,$nillock,getn($nilstacks,$print_class_num)))) in (
	set iheap = heaps($iheap_args);
(*********************************************************************)
	setn(iheap,1) = heap(1,stack0($get_initialized_object(1),$nillock,nilstack0));
	setn(iheap,$input_class_num) = heap(1,$input_stack);
	setn(iheap,$print_class_num) = heap(1,$print_stack);
(*********************************************************************)
      );
    );
  );
);

(***************************************************************)
let loc_sys_in = loc($java_io_inputstream,0)
and loc_sys_out = loc($java_io_printstream,0)
and sys_class_num = mem($java_lang_system,$classes) in (
(*******************************************************)
  let sys_obj = concat(object,(add($sys_class_num,-1),($loc_sys_in,$loc_sys_out))) in (
    set isheap = heaps($initialized_objects);
    setn(isheap,$sys_class_num) = $sys_obj;
(***************************************************)
  );
);
(***************************************************************)

(*******************************************************************)
let istore = threadlist(thread(threadloc(loc($main_class,0),0),$iframe,nilcallstack),nilthreadlist) in (
  let istate = state($istore,$iheap,$isheap,1) in (
    set iostate = IO($istate,var(x),niloutstack);
  );
);
(*******************************************************************)

(* Rewriting of initialJavaState *)
genrule(initialJavaState(var(x)),$iostate);

(*********************************************************************)
genrule(instackExample,instack(3,instack(1,instack(2,
instack(0,instack(-1,nilinstack))))));
(*********************************************************************)

defun get_class(i) =
(
  getn($classes,$i);
);

defun get_field(i) =
(
  getn($fields,$i);
);

defun get_class_method_flags(i,j) =
(
  getn(getn($methods_flags_per_classes,$i),$j);
);

defun gen_heaps(i,a,b) =
(
  heaps(subsn($heap_vars,$i,heap($a,$b)));
);

(* Definition of initialization for Java.Lang.Object *)
genrule(
  state(threadlist(thread(var(tl),frame(name($init_method,$java_lang_object),pp0,nilstack,locals($local_vars)),
  callstack(storedframe(var(m),var(pc),var(s),var(l)),var(f))),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);

(* Definition of initialization for Java.Lang.Thread *)
genrule(
  state(threadlist(thread(var(tl),frame(name($init_method,$java_lang_thread),pp0,nilstack,locals($local_vars)),
  callstack(storedframe(var(m),var(pc),var(s),var(l)),var(f))),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);

(* Scheduling *)
genrule(
  threadlist(thread(var(tl),frame(var(m),var(pc),var(s),var(l)),var(f)),threadlist(var(y),var(z))),
  threadlist(var(y),threadlist(thread(var(tl),frame(var(m),var(pc),var(s),var(l)),var(f)),var(z)))
);
genrule(
  threadlist(thread(var(tl),xframe(xMonitorEnter(xTakeLock(val_false,var(sm))),
  var(m),var(pc),var(s),var(l)),var(f)),threadlist(var(y),var(z))),
  threadlist(var(y),threadlist(thread(var(tl),
  xframe(xMonitorEnter(xTakeLock(val_false,var(sm))),var(m),var(pc),var(s),var(l)),var(f)),var(z)))
);
genrule(
  threadlist(thread(var(tl),xframe(returnVoid,var(m),var(pc),var(s),var(l)),nilcallstack),
  threadlist(var(y),var(z))),
  threadlist(var(y),threadlist(thread(var(tl),xframe(returnVoid,var(m),var(pc),var(s),var(l)),nilcallstack),
  var(z)))
);

(* Getting the class of a reference *)
genrule(classloc(loc(var(x),var(y))),var(x));
