load definitions.rex;
load computations.rex;
load synchronisation.rex;
load java.io.InputStream.rex;
load java.io.PrintStream.rex;
load java.lang.String.rex;
load java.lang.Thread.rex;

(*** Generation of the first steps for all instructions in the program ***)
for i from 1 to $nb_insts
do(
  let inst = getn($insts,$i) in (
    let class_name = getn($inst,1)
    and method_name = getn($inst,2)
    and pc = getn($inst,3)
    and opcode = getn($inst,4) in (
      genrule(
	frame(name($method_name,$class_name),$pc,var(s),var(l)),
	xframe($opcode,name($method_name,$class_name),$pc,var(s),var(l))
      );
    );
  );
);

(***************************)
(* Definition of 'next' operation on instructions *)
for i in $pcs
do(
  genrule(
    next($i),
    add($i,1)
  );
);
(*************************)

(** Instructions **)

(* new *)
(* the JVM specification tells that the heap is set with uninitialized objects *)
(* but we make a simplification so that we won't put initialized objects *)
(* in the <init> method of java.lang.Object *)
for i from 1 to $nb_classes
do(
  let class = $get_class($i)
  and stack = $get_stack($i)
  and object = $get_initialized_object($i) in (
    let stack_object = op($stack,($object,$nillock,var(h))) in (
      genrule(
	state(threadlist(thread(var(tl),xframe(ins_new($class),var(m),var(pc),var(s),var(l)),
	var(f)),var(ntl)),$gen_heaps($i,var(a),var(h)),var(k),var(tc)),
	state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),stack(loc($class,var(a)),var(s)),var(l)),

(*******************************************)
	var(f)),var(ntl)),$gen_heaps($i,plus(var(a),1),$stack_object),var(k),var(tc))
(*******************************************)

      );
    );
  );
);

(* goto *)
genrule(
  xframe(ins_goto(var(x)),var(m),var(pc),var(s),var(l)),
  frame(var(m),var(x),var(s),var(l))
);

(* returnVoid *)
genrule(
  state(threadlist(thread(var(tl),xframe(returnVoid,var(a),var(b),var(c),var(z)),
  callstack(storedframe(var(m),var(pc),var(s),var(l)),var(f))),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);

(* return *)
genrule(
  state(threadlist(thread(var(tl),xframe(ins_return,var(a),var(b),stack(var(x),var(y)),var(c)),
  callstack(storedframe(var(m),var(pc),var(s),var(l)),var(f))),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),stack(var(x),var(s)),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);

(* nop *)
genrule(
  xframe(nop,var(m),var(pc),var(s),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l))
);

(* dup *)
genrule(
  xframe(dup,var(m),var(pc),stack(var(x),var(s)),var(l)),
  frame(var(m),next(var(pc)),stack(var(x),stack(var(x),var(s))),var(l))
);

(* dup2 *)
genrule(
  xframe(dup2,var(m),var(pc),stack(var(x),stack(var(y),var(s))),var(l)),
  frame(var(m),next(var(pc)),stack(var(x),stack(var(y),stack(var(x),stack(var(y),var(s))))),var(l))
);

(* pop *)
genrule(
  xframe(pop,var(m),var(pc),stack(var(x),var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l))
);

(* pop2 *)
genrule(
  xframe(pop2,var(m),var(pc),stack(var(x),stack(var(y),var(s))),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l))
);

(* swap *)
genrule(
  xframe(swap,var(m),var(pc),stack(var(x),stack(var(y),var(s))),var(l)),
  frame(var(m),next(var(pc)),stack(var(y),stack(var(x),var(s))),var(l))
);

(* push const *)
(* generic operator for iconst, lconst, bipush, sipush, ldc *)
genrule(
  xframe(push(var(x)),var(m),var(pc),var(s),var(l)),
  frame(var(m),next(var(pc)),stack(var(x),var(s)),var(l))
);

(* aConstNull *)
genrule(
  xframe(aConstNull,var(m),var(pc),var(s),var(l)),
  frame(var(m),next(var(pc)),stack(val_null,var(s)),var(l))
);







(* ifNull *)
genrule(
  xframe(ifNull(var(x)),var(m),var(pc),stack(loc(var(a),var(y)),var(s)),var(l)),
 frame(var(m),next(var(pc)),var(s),var(l))
);
genrule(
  xframe(ifNull(var(x)),var(m),var(pc),stack(val_null,var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l))
);

(* ifNonNull *)
genrule(
  xframe(ifNonNull(var(x)),var(m),var(pc),stack(loc(var(a),var(y)),var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l))
);
genrule(
  xframe(ifNonNull(var(x)),var(m),var(pc),stack(val_null,var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l))
);





(* load *)
for i from 1 to $max_locals
do(
  let local = getn($locals,$i)
  and locals = locals($local_vars) in (
    let local_var = getn($locals,$i) in (
      genrule(
	xframe(load($local),var(m),var(pc),var(s),$locals),
	frame(var(m),next(var(pc)),stack($local_var,var(s)),$locals)
      );
    );
  );
);

(* store *)
for i from 1 to $max_locals
do(
  let local = getn($locals,$i)
  and locals = locals($local_vars) in (
    genrule(
      xframe(store($local),var(m),var(pc),stack(var(x),var(s)),$locals),
      frame(var(m),next(var(pc)),var(s),subsn($locals,$i,var(x)))
    );
  );
);

(*******************Xadd remplace************)
(* iInc *)
for i from 1 to $max_locals
do(
  let local = getn($locals,$i)
  and locals = locals($local_vars) in (
    let local_var = getn($locals,$i) in (
      genrule(
	xframe(iInc($local,var(x)),var(m),var(pc),var(s),$locals),
	xframe(wait,var(m),var(pc),var(s),subsn($locals,$i,plus($local_var,var(x))))
      );
(********************enlev le result car a posait problme***************)
(********************xframe(wait,var(m),var(pc),var(s),subsn($locals,$i,result(var(x)))),************)
      genrule(
	xframe(wait,var(m),var(pc),var(s),subsn($locals,$i,var(x))),
	frame(var(m),next(var(pc)),var(s),subsn($locals,$i,var(x)))
      );
    );
  );
);
(****************************)

(*neg*)
genrule(
  xframe(neg,var(m),var(pc),stack(var(a),var(s)),var(l)),
  xframe(result(minus(0,var(a))),var(m),var(pc),var(s),var(l))
);


(* add *)
genrule(
  xframe(add,var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  xframe(result(plus(var(a),var(b))),var(m),var(pc),var(s),var(l))
);

(* sub *)
genrule(
  xframe(sub,var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  xframe(result(minus(var(a),var(b))),var(m),var(pc),var(s),var(l))
);

(*mult*)
 genrule(
  xframe(mult,var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  xframe(result(times(var(a),var(b))),var(m),var(pc),var(s),var(l))
); 





(***??**********)
genrule(eqCl(val_null,val_null),val_true);
for i from 1 to $nb_classes
do(
  let class_i = $get_class($i) in (
    genrule(eqCl($class_i,$class_i),val_true);
    genrule(eqCl($class_i,val_null),val_false);
    for j from 1 to $nb_classes
    do(
      if ($i <> $j) then(
	let class_j = $get_class($j) in (
	  genrule(eqCl($class_i,$class_j),val_false);
	);
      );
    );
  );
);


(* ifACmp *)
genrule(
  xframe(ifACmpEq(var(x)),var(m),var(pc),stack(loc(var(b),var(z)),stack(loc(var(a),var(y)),var(s))),var(l)),
  xframe(xIfACmpEq(var(x),eqCl(var(a),var(b)),var(y),var(z)),var(m),var(pc),var(s),var(l))
);
genrulecond(
  xframe(xIfACmpEq(var(x),val_true,var(a),var(b)),var(m),var(pc),var(s),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  eq(var(a),var(b))
);

(************le ne(var(a),var(b)) est remplac par deux gt et lt car filtre + ****************)
genrulecond(
  xframe(xIfACmpEq(var(x),val_true,var(a),var(b)),var(m),var(pc),var(s),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  gt(var(a),var(b))
);
genrulecond(
  xframe(xIfACmpEq(var(x),val_true,var(a),var(b)),var(m),var(pc),var(s),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  lt(var(a),var(b))
);

(*************************************************************)

genrule(
xframe(xIfACmpEq(var(x),val_false,var(a),var(b)),var(m),var(pc),var(s),var(l)),
frame(var(m),next(var(pc)),var(s),var(l))
);


genrule(
  xframe(ifACmpNe(var(x)),var(m),var(pc),stack(loc(var(b),var(z)),stack(loc(var(a),var(y)),var(s))),var(l)),
  xframe(xIfACmpNe(var(x),eqCl(var(a),var(b)),var(y),var(z)),var(m),var(pc),var(s),var(l))
);
genrule(
 xframe(xIfACmpNe(var(x),val_false,var(a),var(b)),var(m),var(pc),var(s),var(l)),
  frame(var(m),var(x),var(s),var(l))
);

(************le ne(var(a),var(b)) est remplac par deux gt et lt car filtre + ****************)
genrulecond(
xframe(xIfACmpNe(var(x),val_true,var(a),var(b)),var(m),var(pc),var(s),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  gt(var(a),var(b))
);

genrulecond(
xframe(xIfACmpNe(var(x),val_true,var(a),var(b)),var(m),var(pc),var(s),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  lt(var(a),var(b))
);

(*************************************************************)

genrulecond(
xframe(xIfACmpNe(var(x),val_true,var(a),var(b)),var(m),var(pc),var(s),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  eq(var(a),var(b))
);






(* ifCmpNe *)
(************le ne(var(a),var(b)) est remplac par deux gt et lt car filtre + ****************)
genrulecond(
  xframe(ifCmpNe(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  gt(var(a),var(b))
);

genrulecond(
  xframe(ifCmpNe(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  lt(var(a),var(b))
);

(*************************************************************)

genrulecond(
  xframe(ifCmpNe(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  eq(var(a),var(b))
);


(* ifCmpEq *)
genrulecond(
  xframe(ifCmpEq(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  eq(var(a),var(b))
);

(************le ne(var(a),var(b)) est remplac par deux gt et lt car filtre + ****************)
genrulecond(
  xframe(ifCmpEq(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  gt(var(a),var(b))
);

genrulecond(
  xframe(ifCmpEq(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  lt(var(a),var(b))
);

(*************************************************************)

(* ifCmpLt *)
genrulecond(
  xframe(ifCmpLt(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  lt(var(a),var(b))
);

genrulecond(
  xframe(ifCmpLt(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  ge(var(a),var(b))
);


(* ifCmpGt *)
genrulecond(
  xframe(ifCmpGt(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  gt(var(a),var(b))
); 


genrulecond(
  xframe(ifCmpGt(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  le(var(a),var(b))
); 


(*ifCmpLe*)
genrulecond(
  xframe(ifCmpLe(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  le(var(a),var(b))
);

genrulecond(
  xframe(ifCmpLe(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  gt(var(a),var(b))
);

(*ifCmpLe*)
genrulecond(
  xframe(ifCmpGe(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  ge(var(a),var(b))
); 


genrulecond(
  xframe(ifCmpGe(var(x)),var(m),var(pc),stack(var(b),stack(var(a),var(s))),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  lt(var(a),var(b))
); 



(* ifNe *)

(********ne remplac*************)
genrulecond(
  xframe(ifNe(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  gt(var(a),0)
);

genrulecond(
  xframe(ifNe(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  lt(var(a),0)
);

(*******************************)

genrulecond(
  xframe(ifNe(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  eq(var(a),0)
);


(* ifEq *)
genrulecond(
  xframe(ifEq(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  eq(var(a),0)
);

(*************************ne remplac***********)
genrulecond(
  xframe(ifEq(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  gt(var(a),0)
);

genrulecond(
  xframe(ifEq(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  lt(var(a),0)
);

(*******************************)

(* ifLt *)
genrulecond(
  xframe(ifLt(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  lt(var(a),0)
);

genrulecond(
  xframe(ifLt(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  ge(var(a),0)
);


(* ifGt *)
genrulecond(
  xframe(ifGt(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  gt(var(a),0)
); 


genrulecond(
  xframe(ifGt(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  le(var(a),0)
); 



(*ifLe*)
genrulecond(
  xframe(ifLe(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  le(var(a),0)
);

genrulecond(
  xframe(ifLe(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  gt(var(a),0)
);

(*ifGe*)
genrulecond(
  xframe(ifGe(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),var(x),var(s),var(l)),
  ge(var(a),0)
); 


genrulecond(
  xframe(ifGe(var(x)),var(m),var(pc),stack(var(a),var(s)),var(l)),
  frame(var(m),next(var(pc)),var(s),var(l)),
  lt(var(a),0)
); 






(* instanceOf *)
(* TODO : subclass is not enough as it does not take in *)
(* account interfaces. *)
genrule(
  xframe(instanceOf(var(c)),var(m),var(pc),stack(loc(var(x),var(y)),var(s)),var(l)),
  xframe(xInstanceOf(subclass(var(x),var(c))),var(m),var(pc),var(s),var(l))
);



(********************************************************)
genrule(xInstanceOf(val_true),result(1));
genrule(xInstanceOf(val_false),result(0));
(********************************************************)





(* xCall *)
genrule(
  state(threadlist(thread(var(tl),xframe(xCall(locals($local_vars),nilstack),var(m),var(pc),var(s),wait),
  callstack(storedframe(var(sf),var(spc),wait,var(sl)),var(f))),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),frame(var(m),var(pc),var(s),locals($local_vars)),
  callstack(storedframe(var(sf),var(spc),nilstack,var(sl)),var(f))),var(ntl)),var(h),var(k),var(tc))
);
genrule(
  state(threadlist(thread(var(tl),xframe(xCall(locals($local_vars),stack(var(a),var(b))),var(m),var(pc),var(s),wait),
  callstack(storedframe(var(sf),var(spc),wait,var(sl)),var(f))),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),frame(var(m),var(pc),var(s),locals($local_vars)),
  callstack(storedframe(var(sf),var(spc),stack(var(a),var(b)),var(sl)),var(f))),var(ntl)),var(h),var(k),var(tc))
);

(* invokeVirtual *)
genrule(
  state(threadlist(thread(var(tl),xframe(invokeVirtual(var(ca),var(cam)),name(var(m),var(cc)),
  var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xInvokeVirtual(accessibleMC(name(var(ca),lookup(var(ca),var(cam))),var(cc)),
  subclass(typeInstance(lookup(var(ca),var(cam)),var(cam),var(s)),var(ca)),
  typeInstance(lookup(var(ca),var(cam)),var(cam),var(s)),var(cc),var(cam),
  lookupIV(typeInstance(lookup(var(ca),var(cam)),var(cam),var(s)),var(cam))),
  name(var(m),var(cc)),var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);
genrule(
  xInvokeVirtual(pprotected,val_true,var(ic),var(cc),var(cam),var(rc)),
  xInvokeVirtual(subclass(var(ic),var(cc)),val_true,var(ic),var(cc),var(cam),var(rc))
);
genrule(
  state(threadlist(thread(var(tl),xframe(xInvokeVirtual(val_true,val_true,var(ic),var(x),var(cam),var(rc)),
  name(var(m),var(cc)),var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xCall(localCall(var(s),var(rc),var(cam)),rmArgs(var(s),var(rc),var(cam))),
  name(var(cam),var(rc)),pp0,nilstack,wait),
  callstack(storedframe(name(var(m),var(cc)),var(pc),wait,var(l)),
  var(f))),var(ntl)),var(h),var(k),var(tc))
);

(* invokeStatic *)
genrule(
  state(threadlist(thread(var(tl),xframe(invokeStatic(var(ca),var(cam)),name(var(m),var(cc)),
  var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xInvokeStatic(lookup(var(ca),var(cam)),var(cam)),name(var(m),var(cc)),
  var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);
genrule(
  state(threadlist(thread(var(tl),xframe(xInvokeStatic(var(rc),var(cam)),name(var(m),var(cc)),
  var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xCall(localCall(var(s),var(rc),var(cam)),rmArgs(var(s),var(rc),var(cam))),
  name(var(cam),var(rc)),pp0,nilstack,wait),
  callstack(storedframe(name(var(m),var(cc)),var(pc),wait,var(l)),
  var(f))),var(ntl)),var(h),var(k),var(tc))
);

(* invokeSpecial *)
genrule(orn(val_true,val_true),val_true);
genrule(orn(val_true,val_false),val_true);
genrule(orn(val_false,val_true),val_false);
genrule(orn(val_false,val_false),val_true);
genrule(
  state(threadlist(thread(var(tl),xframe(invokeSpecial(var(ca),var(cam)),name(var(m),var(cc)),
  var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xInvokeSpecialCC(accessibleMC(name(var(ca),lookup(var(ca),
  var(cam))),var(cc)),subclass(typeInstance(lookup(var(ca),var(cam)),
  var(cam),var(s)),var(ca)),var(cc),var(ca),var(cam),
  typeInstance(lookup(var(ca),var(cam)),var(cam),var(s))),
  name(var(m),var(cc)),var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);
genrule(
  xInvokeSpecialCC(pprotected,val_true,var(cc),var(ca),var(cam),var(ic)),
  xInvokeSpecialCC(subclass(var(ic),var(cc)),val_true,var(cc),var(ca),var(cam),var(ic))
);
genrule(
  xInvokeSpecialCC(val_true,val_true,var(cc),var(ca),var(cam),var(ic)),
  xInvokeSpecial(orn(isInit(var(cam)),subclass(var(cc),lookup(var(ca),var(cam)))),
  var(ca),var(cam),var(ic))
);
genrule(
  state(threadlist(thread(var(tl),xframe(xInvokeSpecial(val_true,var(ca),var(cam),var(ic)),name(var(m),var(cc)),
  var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xCall(localCall(var(s),lookupSp1(var(ca),var(cam)),var(cam)),
  rmArgs(var(s),lookupSp1(var(ca),var(cam)),var(cam))),
  name(var(cam),lookupSp1(var(ca),var(cam))),pp0,nilstack,wait),
  callstack(storedframe(name(var(m),var(cc)),var(pc),wait,var(l)),var(f))),var(ntl)),var(h),var(k),var(tc))
);
genrule(
  state(threadlist(thread(var(tl),xframe(xInvokeSpecial(val_false,var(ca),var(cam),var(ic)),name(var(m),var(cc)),
  var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xCall(localCall(var(s),lookupSp2(var(ic),var(cam)),var(cam)),
  rmArgs(var(s),lookupSp2(var(ic),var(cam)),var(cam))),
  name(var(cam),lookupSp2(var(ic),var(cam))),pp0,nilstack,wait),
  callstack(storedframe(name(var(m),var(cc)),var(pc),wait,var(l)),var(f))),var(ntl)),var(h),var(k),var(tc))
);

(** getputfieldstatic **)

(* getfield *)
genrule(
  state(threadlist(thread(var(tl),xframe(getField(var(fx)),var(m),var(pc),
  stack(var(x),var(s)),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xGetField(lookupF(var(fx)),xGetHeapStack(var(x),var(h))),
  var(m),var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);

let f_used = 0 in (
  for i from 1 to $nb_classes
  do(
    let stack = $get_stack($i)
    and object = $get_object_var($i)
    and nb_fields = getn($nb_fields_per_classes,$i) in (      
      for j from 1 to $nb_fields
      do(
	set f_used = add(1,$f_used);
	let fieldname = getn($fields,$f_used)
	and fieldvar = getn($field_vars,$f_used) in (
	  genrule(
	    xGetField($fieldname,resultstack(var(ds),op($stack,($object,var(lk),var(s))))),
	    resultGetField($fieldvar)
	  );
	);
      );
    );
  );
);

genrule(
  state(threadlist(thread(var(tl),xframe(resultGetField(var(x)),var(m),var(pc),var(s),var(l)),
  var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),stack(var(x),var(s)),var(l)),
  var(f)),var(ntl)),var(h),var(k),var(tc))
);

(* getstatic *)
for ff in $fields_flags
do(
  let field = getn($ff,1)
  and flags = getn($ff,2) in (
    let class = getn($field,1) in (
      let class_num = mem($class,$classes) in (
	let field_num = mem($field,getn($fields_per_classes,$class_num))
	and obj_var = getn($objects_vars,$class_num) in (
	  let f_var = getn($obj_var,$field_num)
	  and heaps_obj = heaps(subsn($heap_vars,$class_num,$obj_var))
	  and is_static = mem(AccStatic,$flags) in (
	    if ($is_static) then
	    (
	      genrule(
		state(threadlist(thread(var(tl),xframe(getStatic($field),var(m),var(pc),var(s),var(l)),
		var(f)),var(ntl)),var(h),$heaps_obj,var(tc)),
		state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),stack($f_var,var(s)),var(l)),
		var(f)),var(ntl)),var(h),$heaps_obj,var(tc))
	      );
	    );
	  );
	);
      );
    );
  );
);

(* putStatic *)
for ff in $fields_flags
do(
  let field = getn($ff,1)
  and flags = getn($ff,2) in (
    let class = getn($field,1) in (
      let class_num = mem($class,$classes) in (
	let field_num = mem($field,getn($fields_per_classes,$class_num))
	and obj_var = getn($objects_vars,$class_num) in (
	  let f_var = getn($obj_var,$field_num)
	  and heaps_obj = heaps(subsn($heap_vars,$class_num,$obj_var))
	  and is_static = mem(AccStatic,$flags) in (
	    if ($is_static) then
	    (
	      genrule(
		state(threadlist(thread(var(tl),xframe(putStatic($field),var(m),var(pc),stack(var(x),var(s)),
		var(l)),var(f)),var(ntl)),var(h),$heaps_obj,var(tc)),
		state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),var(s),var(l)),var(f)),var(ntl)),
		var(h),heaps(subsn($heap_vars,$class_num,subsn($obj_var,$field_num,var(x)))),var(tc))
	      );
	    );
	  );
	);
      );
    );
  );
);

(* putfield *)
genrule(
  state(threadlist(thread(var(tl),xframe(putField(var(fx)),var(m),var(pc),
  stack(var(v),stack(var(x),var(s))),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc)),
  state(threadlist(thread(var(tl),xframe(xPutField(lookupF(var(fx)),xGetHeapStack(var(x),var(h)),var(v)),
  var(m),var(pc),var(s),var(l)),var(f)),var(ntl)),var(h),var(k),var(tc))
);

let f_used = 0 in (
  for i from 1 to $nb_classes
  do(
    let stack = $get_stack($i)
    and object = $get_object_var($i)
    and nb_fields = getn($nb_fields_per_classes,$i) in (      
      for j from 1 to $nb_fields
      do(
	set f_used = add(1,$f_used);
	let fieldname = getn($fields,$f_used)
	and fieldvar = getn($field_vars,$f_used) in (
	  genrule(
	    xPutField($fieldname,resultstack(var(ds),op($stack,($object,var(lk),var(s)))),var(v)),
	    resultPutField(xMergeHeapStack(var(ds),op($stack,(subsn($object,$j,var(v)),var(lk),var(s)))))
	  );
	);
      );
      genrule(
	state(threadlist(thread(var(tl),
	xframe(resultPutField(resultstackmerge(op($stack,(var(o),var(lk),var(hs))))),
	var(m),var(pc),var(s),var(l)),var(f)),var(ntl)),$gen_heaps($i,var(a),var(h)),var(k),var(tc)),
	state(threadlist(thread(var(tl),frame(var(m),next(var(pc)),var(s),var(l)),
	var(f)),var(ntl)),$gen_heaps($i,var(a),op($stack,(var(o),var(lk),var(hs)))),var(k),var(tc))
      );
    );
  );
);

(* subclasses *)
for i from 1 to $nb_classes
do(
  let class_i = getn($classes,$i)
  and subclasses = getn($subclasses_per_classes,$i) in (
    for j from 1 to $nb_classes
    do(
      let class_j = getn($classes,$j) in (
	if (mem($j,$subclasses)) then
	(
	  genrule(subclass($class_j,$class_i),val_true);
	) else
	(
	  genrule(subclass($class_j,$class_i),val_false);
	);
      );
    );
  );
);

(* isInit *)
for i from 1 to $nb_methods
do(
  let method = getn($methods,$i) in (
    if (mem(<<init>>,$method)) then
    (
      genrule(isInit($method),val_true);
    ) else
    (
      genrule(isInit($method),val_false);
    );
  );
);

(* localCall, rmArgs, typeInstance *)
for i from 1 to $nb_classes
do(
  let class = $get_class($i)
  and methods_class = getn($methods_per_classes,$i) in (
    for j from 1 to getn($nb_methods_per_classes,$i)
    do(
      let method = getn($methods_class,$j)
      and method_flags = $get_class_method_flags($i,$j) in (
	let nb_args = symboldepth(getn($method,2))
	and is_static = mem(AccStatic,getn($method_flags,2))
	and is_native = mem(AccNative,getn($method_flags,2))
	and stack = var(s) and locals = $nillocals in (
	  if ($is_static) then
	  (
	    for l from 1 to $nb_args
	    do(
	      set local_var_l = getn($local_vars,$l);
	      set stack = concat(stack,(($local_var_l,$stack)));
	      setn(locals,$l) = $local_var_l;
	    );
	  ) else
	  (
	    if ($is_native) then
	    (
	      set stack = stack(var(l0),var(s));
	      set locals = subsn($nillocals,1,var(l0));
	    ) else
	    (
	      set stack = stack(loc(var(ic),var(x)),var(s));
	      set locals = subsn($nillocals,1,loc(var(ic),var(x)));
	    );
	    
	    for l from 2 to add(1,$nb_args)
	    do(
	      set local_var_l = getn($local_vars,$l);
	      set stack = concat(stack,(($local_var_l,$stack)));
	      setn(locals,$l) = $local_var_l;
	    );
	  );
	  
	  if (not($is_static)) then
	  (
	    if ($is_native) then
	    (
	      genrule(typeInstance($class,$method,var(s)),$class);
	    ) else
	    (
	      genrule(typeInstance($class,$method,$stack),var(ic));
	    );
	  );
	  
	  genrule(localCall($stack,$class,$method),$locals);
	  genrule(rmArgs($stack,$class,$method),var(s));
	);
      );
    );
  );
);

(* accessibleMC *)
(* TODO : correct definition *)
(* This works for a few examples when all classes are Public and in *)
(* the same runtime package *)
genrule(accessibleMC(var(x),var(y)),val_true);

(* lookup *)
for i from 1 to $nb_classes
do(
  let class = $get_class($i)
  and superclasses = getn($superclasses_per_classes,$i)
  and l_done = () in (
    for j from 1 to len($superclasses)
    do(
      let superclass_id = getn($superclasses,$j) in (
	let superclass = $get_class($superclass_id)
	and methods_in_superclass = getn($methods_per_classes,$superclass_id) in (
	  for m in $methods_in_superclass
	  do(
	    let e = ($class,$m) in (
	      if (not(mem($e,$l_done))) then
	      (
		genrule(op(lookup,$e),$superclass);
		set l_done = append($e,$l_done);
	      );
	    );
	  );
	);
      );
    );
  );
);

(* lookupIV *)
for i from 1 to $nb_classes
do(
  let class = $get_class($i)
  and superclasses = getn($superclasses_per_classes,$i)
  and l_done = () in (
    for j from 1 to len($superclasses)
    do(
      let superclass_id = getn($superclasses,$j) in (
	let superclass = $get_class($superclass_id)
	and methods_in_superclass = getn($methods_per_classes,$superclass_id) in (
	  for m in $methods_in_superclass
	  do(
	    let e = ($class,$m)
	    and k = mem($m,$methods_in_superclass) in (
	      let flags = getn($get_class_method_flags($superclass_id,$k),2) in (
		if (not(mem($e,$l_done)) && not(mem(<<init>>,$m))
		&& not(mem(<<clinit>>,$m)) && not(mem(AccStatic,$flags))
		&& not(mem(AccAbstract,$flags))) then
		(
		  genrule(op(lookupIV,$e),$superclass);
		  set l_done = append($e,$l_done);
		);
	      );
	    );
	  );
	);
      );
    );
  );
);

(* lookupSp1 *)
for i from 1 to $nb_classes
do(
  let class = $get_class($i)
  and superclasses = getn($superclasses_per_classes,$i)
  and l_done = () in (
    for j from 1 to len($superclasses)
    do(
      let superclass_id = getn($superclasses,$j) in (
	let superclass = getn($classes,$superclass_id)
	and methods_in_superclass = getn($methods_per_classes,$superclass_id) in (
	  for m in $methods_in_superclass
	  do(
	    let e = ($class,$m)
	    and k = mem($m,$methods_in_superclass) in (
	      let flags = getn($get_class_method_flags($superclass_id,$k),2) in (
		if (not(mem($e,$l_done)) && not(mem(<<init>>,$m))
		&& not(mem(<<clinit>>,$m)) && not(mem(AccStatic,$flags))
		&& not(mem(AccAbstract,$flags))) then
		(
		  genrule(op(lookupSp1,$e),$superclass);
		  set l_done = append($e,$l_done);
		) else
		(
		  if (mem(<<init>>,$m) && $superclass = $class) then
		  (
		    genrule(op(lookupSp1,$e),$superclass);
		    set l_done = append($e,$l_done);
		  );
		);
	      );
	    );
	  );
	);
      );
    );
  );
);

(* lookupSp2 *)
for i from 1 to $nb_classes
do(
  let class = $get_class($i)
  and superclasses = getn($superclasses_per_classes,$i)
  and l_done = () in (
    for j from 1 to len($superclasses)
    do(
      let superclass_id = getn($superclasses,$j) in (
	let superclass = getn($classes,$superclass_id)
	and methods_in_superclass = getn($methods_per_classes,$superclass_id) in (
	  for m in $methods_in_superclass
	  do(
	    let e = ($class,$m)
	    and k = mem($m,$methods_in_superclass) in (
	      let flags = getn($get_class_method_flags($superclass_id,$k),2) in (
		if (not(mem($e,$l_done)) && $superclass <> $class
		&& not(mem(<<init>>,$m))
		&& not(mem(<<clinit>>,$m)) && not(mem(AccStatic,$flags))
		&& not(mem(AccAbstract,$flags))) then
		(
		  genrule(op(lookupSp2,$e),$superclass);
		  set l_done = append($e,$l_done);
		);
	      );
	    );
	  );
	);
      );
    );
  );
);

(* lookupF *)
(* Works, but not a good implementation, fields data structure might be revised *)
for i from 1 to $nb_classes
do(
  let class = $get_class($i)
  and nb_fields = getn($nb_fields_per_classes,$i)
  and fields = getn($fields_per_classes,$i) in (
    for j from 1 to $nb_fields
    do(
      let field = getn($fields,$j) in (
	genrule(lookupF(subsn($field,1,$class)),$field);
      );
    );
  );
);
