open Environment
open Utils
open AtermsParser

let nilindent = String.make 0 ' '
let indent = String.make 3 ' '
let indent_rule = String.make (String.length "rule(") ' '
(**********************************************************)
let indent_rule_cond = String.make (String.length "rulecond(") ' '
(**********************************************************)

let maude_term = "Term"

let screen_width = ref 100

let get_screen_width () = !screen_width
let set_screen_width size = screen_width := size

let prepare_string len s =
  if (String.length s > len) then
    let index = ref (len - 1) in
      while not(!index < 0 || s.[!index] = ',' || s.[!index] = '('
	  || s.[!index] = ' ') do
	index := !index - 1
      done;
      if (!index < 0) then
	((String.sub s 0 len),(String.sub s len
					((String.length s) - len)))
      else
	((String.sub s 0 (!index + 1)),(String.sub s (!index + 1)
					  ((String.length s) - !index - 1)))
  else (s,"")

let rec cut_full_string indent len s =
  let offset = (String.length indent) in
  let index = ref (len + offset - 1) in
    (
      if (String.length s > len + offset) then
	(
	  while not(!index < 0 || s.[!index] = ',' || s.[!index] = '('
	      || s.[!index] = ' ') do
	    index := !index - 1
	  done;
	  if (!index < 0) then
	    (indent ^ (String.sub s 0 len)) ::
	      (cut_full_string indent len (String.sub s len
					((String.length s) - len)))
	  else
	    (indent ^ (String.sub s 0 (!index + 1))) :: 
	      (cut_full_string indent len (String.sub s (!index + 1)
					((String.length s) - !index - 1)))
	)
      else
	[indent ^ s]
    )

let cut_prepared_string indent len s =
  let str = prepare_string len s in
    if (snd str = "") then [fst str]
    else
      (fst str) :: (cut_full_string indent len (snd str))

let get_operator_list l =
  String.concat ","
    (List.map
       (fun x -> match x with
	  | (op,arity) ->
	      "op" ^ encaps(
		op ^ "," ^ (string_of_int (max arity 0))
	      )
       ) l)

let get_renamed_vars ops vars =
  let renamed_vars = ref [] in
    (List.iter
       (fun var ->
	  let renamed_var = resolve_name_conflict var
	    ((fst (List.split ops)) @ (snd (List.split !renamed_vars))) in
	    (if (var <> renamed_var) then
	       let really_renamed_var =
		 if (List.mem renamed_var vars) then
		   let conflict_list =
		     (fst (List.split ops)) @ vars
		     @ (snd (List.split !renamed_vars)) in
		     resolve_name_conflict renamed_var conflict_list
		 else renamed_var in
		 renamed_vars := (var,really_renamed_var) :: !renamed_vars
	     else ())) vars);
    !renamed_vars

let rename_vars s renamed_vars =
  List.fold_left
    (fun x y -> Str.global_replace
       (Str.regexp ("var(" ^ y ^ ")")) ("var(" ^ (List.assoc y renamed_vars) ^ ")") x)
    s (fst (List.split renamed_vars))

let get_program_ops ops =
  fst (List.split (List.find_all (fun x -> (fst x).[0] = '<') ops))

let is_program_op op = (op.[0] = '<')

let rename_program_op op ops vars =
  let renamed_op = resolve_name_conflict (decaps_prog op)
    ((fst (List.split ops)) @ vars) in renamed_op

let rename_program_ops s ops vars = 
  let prog_ops = get_program_ops ops in
    List.fold_left
      (fun x y -> Str.global_replace
	 (Str.regexp y) (rename_program_op y ops vars) x)
      s prog_ops
        
let simplify_vars s =
  Str.global_replace
    (Str.regexp ("var(\\([a-zA-Z0-9_-]+\\))")) "\\1" s

(**************************************************************************)
let get_renamed_system ops vars rules rules_cond=
  let renamed_vars = get_renamed_vars ops vars in
  let new_vars = List.map (fun var ->
			     if (List.mem_assoc var renamed_vars) then
			       List.assoc var renamed_vars
			     else var) vars in
  let rename_ops s = rename_program_ops s ops new_vars in
  let new_ops = List.map (fun op ->
			    match op with
			      | (n,a) ->
				  if (is_program_op n) then
				    (rename_program_op n ops new_vars,a)
				  else (n,a)) ops in
  let new_rules = List.map (fun r ->
			      match r with
				| (rl, rr) ->
				    (rename_ops (rename_vars rl renamed_vars),
				     rename_ops (rename_vars rr renamed_vars))) rules in
(*****************************************************)
  let new_rules_cond = List.map (fun r ->
			      match r with
				| (rl, (rr, cd)) ->
				    (rename_ops (rename_vars rl renamed_vars),
				     (rename_ops (rename_vars rr renamed_vars),
				     rename_ops (rename_vars cd renamed_vars))))
    rules_cond
(*****************************************************)
  in (new_ops,new_vars,new_rules,new_rules_cond)

(**************************************************************************)


let get_maude_operator_signature ops =
  List.map
    (fun op ->  match op with
       | (n,a) -> let get_list =
	   let l = ref [] in
	     for i = 1 to a do
	       l := 0 :: !l
	     done; !l in
	   "op " ^ (maude_replace_underscores n) ^ " : " ^
	     (String.concat " " (List.map (fun _ -> maude_term) get_list)) ^
	     " -> " ^ maude_term ^ " .") ops

let get_timbuk_operator_signature ops =
  List.map
    (fun op ->  match op with
       | (n,a) ->
	   (timbuk_epure n) ^ ":" ^ (string_of_int a)) ops

let get_maude_operator_list ops =
  String.concat "\n" (get_maude_operator_signature ops)

let get_timbuk_operator_list ops =
  "Ops " ^ (String.concat " " (get_timbuk_operator_signature ops))

let get_var_list l =
  String.concat "," l

let get_maude_var_list l =
  "vars " ^ (String.concat " " (List.map maude_replace_underscores l))
  ^ " : " ^ maude_term ^ " ."

let get_timbuk_var_list l =
  "Vars " ^ (String.concat " " (List.map timbuk_epure l))

let get_trs_list l=
if (l = []) then [""]
else
  let rev_l = List.rev l  in
    (List.flatten(List.map
		    (fun x -> match x with
		       | (sb,se) ->
			   if (String.length (sb ^ se ^ indent ^ indent_rule)
			       < get_screen_width()) then
			     [ "rule(" ^ sb ^ "," ^ se ^ ")," ]
			   else
			     (cut_prepared_string (indent ^ indent_rule)
				(get_screen_width()) ("rule(" ^ sb ^ ",")) @
			       (cut_prepared_string
				  (indent ^ indent_rule) (get_screen_width())
				  (indent_rule ^ se ^ "),"))
		    )
		    (List.rev(List.tl rev_l)))) @

      (match List.hd rev_l with
	 | (sb,se) ->
	     if (String.length (sb ^ se ^ indent ^ indent_rule)
		 < get_screen_width()) then
	       [ "rule(" ^ sb ^ "," ^ se ^ ")" ]
	     else
	       (cut_prepared_string (indent ^ indent_rule)
		  (get_screen_width()) ("rule(" ^ sb ^ ",")) @
		 (cut_prepared_string
		    (indent ^ indent_rule) (get_screen_width())
		    (indent_rule ^ se ^ ")"))
      ) 


(***********************************************************************************)
let get_trs_cond_list l=
if (l = []) then [""]
else
  let rev_l = List.rev l in
    (List.flatten(List.map
		    (fun x -> match x with
		       | (sb,(se,sc)) ->
			   if (String.length (sb ^ se ^ sc ^ indent ^ indent_rule_cond)
			       < get_screen_width()) then
			     [ "rulecond(" ^ sb ^ "," ^ se ^ "," ^ sc ^ ")," ]
			   else
			     (cut_prepared_string (indent ^ indent_rule_cond)
				(get_screen_width()) ("rulecond(" ^ sb ^ ",")) @
			       (cut_prepared_string
				  (indent ^ indent_rule_cond) (get_screen_width())
				  (indent_rule_cond ^ se ^ ",")) @
			       (cut_prepared_string
				  (indent ^ indent_rule_cond) (get_screen_width())
				  (indent_rule_cond ^ sc ^ "),"))
		    )
		    (List.rev(List.tl rev_l)))) @
      (match List.hd rev_l with
	 | (sb,(se,sc)) ->
	     if (String.length (sb ^ se ^ sc ^ indent ^ indent_rule_cond)
		 < get_screen_width()) then
	       [ "rulecond(" ^ sb ^ "," ^ se ^  "," ^ sc ^")" ]
	     else
	       (cut_prepared_string (indent ^ indent_rule_cond)
		  (get_screen_width()) ("rulecond(" ^ sb ^ ",")) @
		 (cut_prepared_string
		    (indent ^ indent_rule_cond) (get_screen_width())
		    (indent_rule_cond ^ se ^ ","))@
		 (cut_prepared_string
		    (indent ^ indent_rule_cond) (get_screen_width())
		    (indent_rule_cond ^ sc ^ ")"))

      ) 
(********************************************************************************)



let get_maude_trs_list l =
  (List.flatten(List.map
		  (fun x -> match x with
		     | (rl,rr) ->
			 let (rl,rr) =
			   (maude_replace_underscores (simplify_vars rl),
			    maude_replace_underscores (simplify_vars rr))
			 in
			   if (String.length (rl ^ rr ^ indent)
			       < get_screen_width()) then
			     [ "rl " ^ rl ^ " => " ^ rr ^ " ." ]
			   else
			     (cut_prepared_string (indent)
				(get_screen_width()) ("rl " ^ rl ^ " => ")) @
			       (cut_prepared_string
				  (indent) (get_screen_width())
				  (indent ^ rr ^ " ."))
		  )
		  l)
  )


(*************************************************************************************)



let get_maude_trs_cond_list l =
  (List.flatten(List.map
		  (fun x -> match x with
		     | (rl,(rr,cd)) ->
			 let (rl,(rr,cd)) =
			   (maude_replace_underscores (simplify_vars rl),
			    (maude_replace_underscores (simplify_vars rr),
			     maude_replace_underscores (simplify_vars cd)))
		   in 

			   if (String.length (rl ^ rr ^ cd ^ indent)
			       < get_screen_width()) then

			     [ "crl " ^ rl ^ " => " ^ rr ^ " if (" ^ cd ^") ." ]

			   else
			     (cut_prepared_string (indent)
				(get_screen_width()) ("crl " ^ rl ^ " => ")) @
			       (cut_prepared_string
				  (indent) (get_screen_width())
				  (indent ^ rr ^ " if " ^ cd ^ " ."))
		  )
		  l)
  )
(*************************************************************************************)





let get_timbuk_trs_list l =
  (List.flatten(List.map
		  (fun x -> match x with
		     | (rl,rr) ->
			 let (rl,rr) =
			   (timbuk_epure (simplify_vars rl),
			    timbuk_epure (simplify_vars rr))
			 in
			   if (String.length (rl ^ rr ^ indent)
			       < get_screen_width()) then
			     [ rl ^ " -> " ^ rr ]
			   else
			     (cut_prepared_string (indent)
				(get_screen_width()) (rl ^ " -> ")) @
			       (cut_prepared_string
				  (indent) (get_screen_width())
				  (indent ^ rr))
		  )
		  l)
  )




(*************************************************************************************)
let get_timbuk_trs_cond_list l =
  (List.flatten(List.map
		  (fun x -> match x with
		     | (rl,(rr,cd)) ->
			 let (rl,(rr,cd)) =
			   (timbuk_epure (simplify_vars rl),
			    (timbuk_epure (simplify_vars rr),
			     (simplify_vars cd)))
		   in 

			   if (String.length (rl ^ rr ^ cd ^ indent)
			       < get_screen_width()) then
			     [ rl ^ " -> " ^ rr ^ " if " ^ cd ]
			   else
			     (cut_prepared_string (indent)
				(get_screen_width()) (rl ^ " -> ")) @
			       (cut_prepared_string
				  (indent) (get_screen_width())
				  (indent ^ rr ^ "if " ^ cd ))
		  )
		  l)
  )
(*************************************************************************************)



let export env file sysname =
(***let _ = print_string "Export aterm\n" in**)
  let (new_ops, new_vars, new_rules, new_rules_cond) =
    get_renamed_system (env.get_ops()) (env.get_vars()) (env.get_rules()) (env.get_rules_cond()) in
 
  let oc = IO.output_channel (open_out file)
  and ops = get_operator_list new_ops
  and vars = get_var_list new_vars
  and trs = get_trs_list new_rules 
  and trs_cond = get_trs_cond_list new_rules_cond in
    (
      IO.nwrite oc "specification(\n";
      IO.nwrite oc (indent ^ "operatorList([\n");
      List.iter
	(fun x -> IO.nwrite oc (indent ^ indent); IO.nwrite oc x;
	   IO.write oc '\n') (cut_full_string nilindent (get_screen_width())
				(ops ^ "]),"));
      IO.nwrite oc (indent ^ "varList([\n");
      List.iter
	(fun x -> IO.nwrite oc (indent ^ indent); IO.nwrite oc x;
	   IO.write oc '\n') (cut_full_string nilindent (get_screen_width())
				(vars ^ "]),"));
      IO.nwrite oc (indent ^ "trsList([system(\"" ^ sysname ^ "\",[\n");

      let rev_trs = List.rev trs in
	(List.iter
	   (fun x -> IO.nwrite oc (indent ^ indent); IO.nwrite oc x;
	      IO.write oc '\n') (List.rev (List.tl rev_trs));
	 IO.nwrite oc (indent ^ indent ^ (List.hd rev_trs))
	);
       IO.nwrite oc "])]),\n";
(*************************************************************************)
      IO.nwrite oc (indent ^ "trsListCond([system(\"" ^ sysname ^ "\",[\n");

      let rev_trs_cond = List.rev trs_cond in
	(List.iter
	   (fun x -> IO.nwrite oc (indent ^ indent); IO.nwrite oc x;
	      IO.write oc '\n') (List.rev (List.tl rev_trs_cond));
	 IO.nwrite oc (indent ^ indent ^ (List.hd rev_trs_cond))
	);
(*************************************************************************)
      IO.nwrite oc "])])\n)";
      (***let _= print_string "Fin export aterm !!!\n" in ***)
IO.close_out oc
    )

let export_maude aterm_file sysname maude_file =
 (**** let _ = print_string "Export maude\n" in***)
  let aterms_spec = parse_specification aterm_file in
  let aterms_ops = get_aterms_ops aterms_spec
  and aterms_vars = get_aterms_vars aterms_spec
  and aterms_system = get_aterms_system aterms_spec sysname 
(*************************************************************************)
  and aterms_system_cond = get_aterms_system_cond aterms_spec sysname in
(*************************************************************************)
  let oc = IO.output_channel (open_out maude_file)
  and ops = get_maude_operator_list aterms_ops
  and vars = get_maude_var_list aterms_vars
  and trs = get_maude_trs_list aterms_system
(*************************************************************************)
  and trs_cond = get_maude_trs_cond_list aterms_system_cond in
(*************************************************************************)
    ( IO.nwrite oc ("load opLTA.maude .\n");
      IO.nwrite oc ("mod " ^ sysname ^ " is\n");
      IO.nwrite oc ("protecting OPTERM .\n");
      IO.nwrite oc (ops ^ "\n");
      IO.nwrite oc (vars ^ "\n");
      (List.iter
	 (fun x -> IO.nwrite oc x; IO.write oc '\n') trs);
(*************************************************************************)
 (List.iter
	 (fun x -> IO.nwrite oc x; IO.write oc '\n') trs_cond);
(*************************************************************************)
      IO.nwrite oc "endm";
(***let _ = print_string "Fin export maude !!!\n" in***)
      IO.close_out oc
    )

let export_timbuk aterm_file sysname timbuk_file =
(***let _ = print_string "Export timbuk\n" in***)
  let aterms_spec = parse_specification aterm_file in
  let aterms_ops = get_aterms_ops aterms_spec
  and aterms_vars = get_aterms_vars aterms_spec
  and aterms_system = get_aterms_system aterms_spec sysname
(*************************************************************************)
  and aterms_system_cond = get_aterms_system_cond aterms_spec sysname in
(*************************************************************************)
  let oc = IO.output_channel (open_out timbuk_file)
  and ops = get_timbuk_operator_list aterms_ops
  and vars = get_timbuk_var_list aterms_vars
  and trs = get_timbuk_trs_list aterms_system 
(*************************************************************************)
  and trs_cond = get_timbuk_trs_cond_list aterms_system_cond in
(*************************************************************************)
    (
      List.iter
	(fun x -> IO.nwrite oc x;
	   IO.write oc '\n')
	(cut_prepared_string indent (get_screen_width()) ops);
      IO.nwrite oc "\n";
      IO.nwrite oc (vars ^ "\n\n");
      IO.nwrite oc ("TRS " ^ sysname ^ "\n");
      (List.iter
	 (fun x -> IO.nwrite oc x; IO.write oc '\n') trs);
(*************************************************************************)
      (List.iter
	 (fun x -> IO.nwrite oc x; IO.write oc '\n') trs_cond);
(*************************************************************************)
      IO.nwrite oc ("Set Init\n");
      IO.nwrite oc ("initialJavaState(instackExample)\n\n");
(****let _ = print_string "Fin export timbuk !!!\n" in***)
      IO.close_out oc
    )


















