open JClass
open JBasics
open JClassAbs
open Rename
open Terms_of_integer
open Terms_of_string
open Utils
open Stubs

let array_class_name = "NativeArray"
let get_otype_name t =
  let name =
    (match t with
       | TClass cn -> cn2s cn
       | TArray _ -> array_class_name
    ) in name

(* converts an instruction to a term *)

(********************************************)
let inst_to_string inst =
  match inst with
    | OpNop -> "nop"
    | OpConst `ANull -> "aConstNull"
    | OpConst `Int int32 -> "push("^(Int32.to_string int32)^")"
    | OpConst `Long int64 -> "push("^(Int64.to_string int64)^")"
    | OpConst `Byte i -> "push("^(string_of_int i)^")"
    | OpConst `Short i -> "push("^(string_of_int i)^")"
    | OpConst `Class _ -> failwith "Gen.inst_to_string : Opconst `Class t not implemented"
    | OpConst `String s -> "push("^(term_of_string s)^")"
(**********************************************)
    | OpStore (_, i) -> "store(local"^(string_of_int i)^")"
    | OpLoad (_, i) -> "load(local"^(string_of_int i)^")"
    | OpNew cn -> "ins_new("^(cn2s cn)^")"
    | OpPop -> "pop"
    | OpPop2 -> "pop2"
    | OpDup -> "dup"
    | OpDupX1 -> "dupX1"
    | OpDupX2 -> "dupX2"
    | OpDup2 -> "dup2"
    | OpDup2X1 -> "dup2X1"
    | OpDup2X2 -> "dup2X2"
    | OpSwap -> "swap"
    | OpAdd _ -> "add"
    | OpSub _ -> "sub"
    | OpMult _ -> "mult"
    | OpDiv _ -> "div"
    | OpRem _ -> "rem"
    | OpNeg _ -> "neg"
    | OpIShl -> "ishl"
    | OpLShl -> "lshl"
    | OpIShr -> "ishr"
    | OpLShr -> "lshr"
    | OpIUShr -> "iushr"
    | OpLUShr -> "lushr"
    | OpIAnd -> "iand"
    | OpLAnd -> "land"
    | OpIOr -> "ior"
    | OpLOr -> "lor"
    | OpIXor -> "ixor"
    | OpLXor -> "lxor"
    | OpL2I -> "l2i"
    | OpI2L -> "i2l"
    | OpI2B -> "i2b"
    | OpI2C -> "i2c"
    | OpI2S -> "i2s"
    | OpIf (`Ne,i) -> "ifNe(pp"^(string_of_int i)^")"
    | OpIf (`Eq,i) -> "ifEq(pp"^(string_of_int i)^")"
    | OpIf (`Lt,i) -> "ifLt(pp"^(string_of_int i)^")"
    | OpIf (`Ge,i) -> "ifGe(pp"^(string_of_int i)^")"
    | OpIf (`Gt,i) -> "ifGt(pp"^(string_of_int i)^")"
    | OpIf (`Le,i)  -> "ifLe(pp"^(string_of_int i)^")"
    | OpIf (`Null,i) -> "ifNull(pp"^(string_of_int i)^")"
    | OpIf (`NonNull,i) -> "ifNonNull(pp"^(string_of_int i)^")"
    | OpIfCmp (`INe,i) -> "ifCmpNe(pp"^(string_of_int i)^")"
    | OpIfCmp (`IEq,i) -> "ifCmpEq(pp"^(string_of_int i)^")"
    | OpIfCmp (`ILt,i) -> "ifCmpLt(pp"^(string_of_int i)^")"
    | OpIfCmp (`IGe,i) -> "ifCmpGe(pp"^(string_of_int i)^")"
    | OpIfCmp (`IGt,i) -> "ifCmpGt(pp"^(string_of_int i)^")"
    | OpIfCmp (`ILe,i) -> "ifCmpLe(pp"^(string_of_int i)^")"
    | OpIfCmp (`AEq,i) -> "ifACmpEq(pp"^(string_of_int i)^")"
    | OpIfCmp (`ANe,i) -> "ifACmpNe(pp"^(string_of_int i)^")"
    | OpGoto i -> "ins_goto(pp"^(string_of_int i)^")"
    | OpReturn `Void -> "returnVoid"
    | OpReturn _ -> "ins_return"
    | OpGetField (cn,fs) -> "getField("^(fn2s cn fs)^")"
    | OpGetStatic (cn,fs) -> "getStatic("^(fn2s cn fs)^")"
    | OpPutStatic (cn,fs) -> "putStatic("^(fn2s cn fs)^")"
    | OpPutField (cn,fs) -> "putField("^(fn2s cn fs)^")"
    | OpInvoke ((`Special cn),ms) -> "invokeSpecial("^(cn2s cn)^","^(mn2s ms)^")"
    | OpInvoke ((`Virtual t),ms) ->
	let name = get_otype_name t in"invokeVirtual("^name^","^(mn2s ms)^")"
    | OpInvoke ((`Interface cn),ms) -> "invokeInterface("^(cn2s cn)^","^(mn2s ms)^")"
    | OpInvoke ((`Static cn),ms) -> "invokeStatic("^(cn2s cn)^","^(mn2s ms)^")"
    | OpInstanceOf t ->
	let name = get_otype_name t in "instanceOf("^name^")"
    | OpIInc (l,i) -> "iInc(local"^(string_of_int l)^","^(string_of_int i)^")"
    | OpMonitorEnter -> "monitorenter"
    | OpMonitorExit -> "monitorexit"
    | OpThrow -> "throw"
    | x -> failwith
	      ("Error: unsupported instruction "^JDump.opcode x^".")
	

let gen_cmi (ic : interface_or_class) =
  let cns = cn2s(get_name ic)
  and l = ref [] in
    iter_concrete_methods
      (fun ms cm ->
	 let mns = mn2s ms
	 and impl = cm.cm_implementation in
	   match impl with
	     | Native -> ()
	     | Java i ->
		 let code = Lazy.force i in
		 let pc = ref 0 in
		   while (!pc < Array.length code.c_code
			  && code.c_code.(!pc) <> OpInvalid) do
		     l := (cns,mns,"pp" ^ (string_of_int !pc),
			   inst_to_string code.c_code.(!pc)) :: !l;
		     pc := !pc + 1
		   done
      ) ic; List.rev !l

(* generates terms for program instructions *)
let gen (prog : program) =
  List.flatten (List.map gen_cmi prog.p_structure)

let rec find_class (prog: program) (class_name: class_name) : interface_or_class =
  try
    let name = String.concat "." class_name in
      if (List.mem (name) (get_stubs_names())) then
	get_stub name
    else
      List.find (fun ic -> get_name ic = class_name) prog.p_structure
  with
    | Not_found ->
	failwith ("Impossible to find class "^(String.concat "." class_name))

let get_defined_fields (ic : interface_or_class) =
  let l = ref [] in
    iter_fields
      (fun fs f -> l := ((get_name ic),fs,f) :: !l) ic; List.rev !l

let rec get_super_fields (prog : program) (ic_todo : interface_or_class list)
    (cn_done : class_name list) f_done =
  let cnl2icl cnl = List.map (find_class prog) cnl in
    match ic_todo with
      | [] -> f_done
      | ic :: tl ->
	  let name = get_name ic in
	    if (List.mem name cn_done) then
	      get_super_fields prog tl cn_done f_done
	    else
	      let fields = (get_defined_fields (find_class prog name)) in
		(match ic with
		   | `Class c ->
		       (match (c.c_super_class,c.c_interfaces) with
			  | (None,il) ->
			      get_super_fields prog ((cnl2icl il) @ tl)
				(name :: cn_done) (f_done @ fields)
			  | (Some cn,il) ->
			      get_super_fields prog
				((find_class prog cn) :: (cnl2icl il) @ tl)
				(name :: cn_done) (f_done @ fields)
		       )
		   | `Interface i ->
		       get_super_fields prog ((cnl2icl i.i_interfaces) @ tl)
			 (name :: cn_done) (f_done @ fields)
	      )
	  
let get_herited_fields (prog : program) (ic : interface_or_class) =
  get_super_fields prog [ic] [] []
		 
let get_methods (ic : interface_or_class) =
  let l = ref [] in
    iter_methods (fun ms m -> l := (ms,m) :: !l) ic; List.rev !l

let get_classes (prog : program) =
  ExtList.List.filter_map
       (fun x -> match x with `Class x -> Some (`Class x) | _ -> None) prog.p_structure

let get_nb_classes (prog : program) : int =
  List.length (get_classes prog)

let get_nb_fields_per_classes (prog : program) : int list =
  List.map (fun x -> List.length(get_herited_fields prog x))
    (get_classes prog)

let get_nb_methods_per_classes (prog : program) : int list =
  List.map (fun x -> List.length(get_methods x))
    (get_classes prog)

let get_fields_per_classes (prog : program) : string list list =
  List.map (fun flist ->
	      List.map
		(fun x ->
		   let (fn,fs,_) = x in fn2s fn fs)
		(get_herited_fields prog flist))
    (get_classes prog)

let get_nb_fields (prog : program) : int =
  List.length (List.flatten (get_fields_per_classes prog))

let get_methods_per_classes (prog : program) : string list list =
  List.map (fun mlist ->
	      List.map
		(fun x ->
		   let (ms,_) = x in mn2s ms)
		(get_methods mlist))
    (get_classes prog)

let get_nb_methods (prog : program) : int =
  List.length (remove_list_duplicates
		 (List.flatten (get_methods_per_classes prog)))

let get_init_fields_per_classes (prog : program) =
  List.map (fun flist ->
	      List.map
		(fun x ->
		   let (_,fs,_) = x in type2init fs.fs_type)
		(get_herited_fields prog flist))
    (get_classes prog)
      
let get_classes_names (prog : program) =
  List.map (fun x -> cn2s (get_name x))
    (get_classes prog)

let get_fields_names (prog : program) =
  List.flatten (get_fields_per_classes prog)

let get_methods_names (prog : program) =
  remove_list_duplicates
    (List.flatten (get_methods_per_classes prog))

(* [is_subclass p c1 c2] returns true iff [c1] is a sub class of [c2],
    or [c2] itself, in the program [p]. *)
let rec is_subclass (p : program) (c1 : jclass) (c2 : jclass) =
  let trans super = 
    match super with 
      | None -> false
      | Some cn ->
	  let ic = (find_class p cn) in
	    (match ic with
	      | `Class c -> is_subclass p c c2
	      | _ -> failwith "Error : a superclass has been resolved as an interface ??")
  in c1 == c2 || trans c1.c_super_class

let compare_classes (p : program) c1 c2 =
  if (is_subclass p c1 c2) then -1
  else
    if (is_subclass p c2 c1) then 1
    else 0

let get_subclasses_per_classes (p : program) =
  let classes = List.map (fun (`Class x) -> x) (get_classes p) in
    List.map
      (fun c ->
	 let subclass_list = ref []
	 and cur_class = ref 0 in
	   while (!cur_class < List.length classes) do
	     (
	       let subclass = (List.nth classes !cur_class) in
		 if (is_subclass p subclass c) then
		   subclass_list := (!cur_class + 1) :: !subclass_list
		 else ()
	     );
	     cur_class := !cur_class + 1
	   done;
	   List.sort
	     (fun c1 c2 -> compare_classes p (List.nth classes (c2-1))
		(List.nth classes (c1-1))) !subclass_list
      ) classes

let get_superclasses_per_classes (p : program) =
  let classes = List.map (fun (`Class x) -> x) (get_classes p) in
    List.map
      (fun c ->
	 let superclass_list = ref []
	 and cur_class = ref 0 in
	   while (!cur_class < List.length classes) do
	     (
	       let superclass = (List.nth classes !cur_class) in
		 if (is_subclass p c superclass) then
		   superclass_list := (!cur_class + 1) :: !superclass_list
		 else ()
	     );
	     cur_class := !cur_class + 1
	   done;
	   List.sort
	     (fun c1 c2 -> compare_classes p (List.nth classes (c1-1))
		(List.nth classes (c2-1))) !superclass_list
      ) classes

let get_field_flags f =
  let l = ref [] in
    (match f with
       | `ClassField cf ->
	   (match cf.cf_access with
	      | `Default -> l := "AccDefault" :: !l
	      | `Private -> l := "AccPrivate" :: !l
	      | `Protected -> l := "AccProtected" :: !l
	      | `Public -> l := "AccPublic" :: !l
	   );
	   (match cf.cf_kind with
	      | NotFinal -> ()
	      | Final -> l := "AccFinal" :: !l
	      | Volatile ->  l := "AccVolatile" :: !l
	   );
	   if (cf.cf_static) then  l := "AccStatic" :: !l;
	   if (cf.cf_transient) then  l := "AccTransient" :: !l;
       | `InterfaceField _ ->
	   l := "AccInterface" :: !l
    ); !l

let get_class_flags c = 
  let l = ref [] in
    (match c.c_access with
       | `Default -> l := "AccDefault" :: !l
       | `Public -> l := "AccPublic" :: !l
    );
    if (c.c_abstract) then l := "AccAbstract" :: !l;
    if (c.c_final) then l := "AccAbstract" :: !l;
    !l

let get_method_flags m =
  let l = ref [] in
    (match m with
       | AbstractMethod am ->
	   (match am.am_access with
	      | `Default -> l := "AccDefault" :: !l
	      | `Protected -> l := "AccProtected" :: !l
	      | `Public -> l := "AccPublic" :: !l
	   );
	   l := "AccAbstract" :: !l
       | ConcreteMethod cm ->
	   (match cm.cm_access with
	      | `Default -> l := "AccDefault" :: !l
	      | `Private -> l := "AccPrivate" :: !l
	      | `Protected -> l := "AccProtected" :: !l
	      | `Public -> l := "AccPublic" :: !l
	   );
	   if (cm.cm_static) then l := "AccStatic" :: !l;
	   if (cm.cm_final) then l := "AccFinal" :: !l;
	   if (cm.cm_synchronized) then l := "AccSynchronized" :: !l;
	   if (cm.cm_strict) then l := "AccStrict" :: !l;
	   (* a little hack due to the fact that all stub methods have *)
	   (* cm_implementation set to Native even if they are not native... *)
	   if (List.mem 256 cm.cm_other_flags) then l := "AccNative" :: !l
    );
    !l

let get_fields_flags_per_classes (prog : program) =
  List.map (fun flist ->
	      List.map
		(fun x ->
		   let (fn,fs,ff) = x in (fn2s fn fs,get_field_flags ff))
		(get_herited_fields prog flist))
    (get_classes prog)

let get_fields_flags (prog : program) =
  List.flatten (get_fields_flags_per_classes prog)

let get_methods_flags_per_classes (prog : program) =
  List.map (fun mlist ->
	      List.map
		(fun x ->
		   let (ms,m) = x in (mn2s ms,get_method_flags m))
		(get_methods mlist))
    (get_classes prog)

let get_classes_flags (prog : program) =
  let classes = List.map (fun (`Class x) -> x) (get_classes prog) in
    List.map get_class_flags classes
