open Gen
open JClassAbs
open Utils

let gen_inst_set name value =
  "set " ^ name ^ "=" ^ value
    
let program = ref nilprogram
let get_program () = !program
let set_program prog = program := prog
  
let import_java_bytecode () =
  let prog = get_program() in
  let gen_insts = ref [] in
  let insts = gen prog in
    (
      gen_insts :=
	(
	  gen_inst_set "insts"
	    (
	      (encaps(String.concat ","
			(List.map (fun x -> match x with | (a,b,c,d) ->
				     (encaps(String.concat "," [a;b;c;d]))
				  ) insts
			)
		     )
	      )
	    )
	) :: !gen_insts
    );
    (
      gen_insts := (gen_inst_set "nb_insts"
		      (string_of_int (List.length insts))) :: !gen_insts;
      gen_insts := (gen_inst_set "max_locals"
		      (string_of_int prog.p_max_locals)) :: !gen_insts;
      gen_insts := (gen_inst_set "max_pc" (string_of_int (prog.p_max_pcs - 1)))
      :: !gen_insts;
      gen_insts := (gen_inst_set "nb_classes"
		      (string_of_int (get_nb_classes prog)))
      :: !gen_insts;
      gen_insts := (gen_inst_set "nb_fields"
		      (string_of_int(get_nb_fields prog)))
      :: !gen_insts;
      gen_insts := (gen_inst_set "nb_methods"
		      (string_of_int(get_nb_methods prog)))
      :: !gen_insts;
      gen_insts := (gen_inst_set "nb_fields_per_classes"
		      (encaps(String.concat ","
				(List.map string_of_int
				   (get_nb_fields_per_classes prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "nb_methods_per_classes"
		      (encaps(String.concat ","
				(List.map string_of_int
				   (get_nb_methods_per_classes prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "classes"
		      (encaps(String.concat ","
				(get_classes_names prog))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "fields"
		      (encaps(String.concat ","
				(get_fields_names prog))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "methods"
		      (encaps(String.concat ","
				(get_methods_names prog))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "fields_per_classes"
		      (encaps(String.concat ","
				(List.map
				   (fun x ->
				      encaps(String.concat  ","
					       (match x with
						  | [] -> [""]
						  | _ -> x)))
				   (get_fields_per_classes prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "methods_per_classes"
		      (encaps(String.concat ","
				(List.map
				   (fun x -> 
				      encaps(String.concat  ","
					       (match x with
						  | [] -> [""]
						  | _ -> x)))
				   (get_methods_per_classes prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "init_fields_per_classes"
		      (encaps(String.concat ","
				(List.map
				   (fun x -> 
				      encaps(String.concat  ","
					       (match x with
						  | [] -> [""]
						  | _ -> x)))
				   (get_init_fields_per_classes prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "subclasses_per_classes"
		      (encaps(String.concat ","
				(List.map
				   (fun x -> 
				      encaps(String.concat  ","
					       (match x with
						  | [] -> [""]
						  | _ -> List.map string_of_int x)))
				   (get_subclasses_per_classes prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "superclasses_per_classes"
		      (encaps(String.concat ","
				(List.map
				   (fun x -> 
				      encaps(String.concat  ","
					       (match x with
						  | [] -> [""]
						  | _ -> List.map string_of_int x)))
				   (get_superclasses_per_classes prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "classes_flags"
		      (encaps(String.concat ","
				(List.map
				   (fun x -> 
				      encaps(String.concat  ","
					       (match x with
						  | [] -> [""]
						  | _ -> x)))
				   (get_classes_flags prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "fields_flags_per_classes"
		      (encaps(String.concat ","
				(List.map
				   (fun x ->
				      match x with
					| [] -> "()"
					| _ ->
					    encaps(
					      String.concat ","
						(List.map
						   (fun (f,l) ->
						      encaps(f ^ ","
							     ^ encaps(String.concat "," l))
						   ) x)
					    )
				   )
				   (get_fields_flags_per_classes prog)))))
      :: !gen_insts;
      gen_insts := (gen_inst_set "fields_flags"
		      (encaps(
			 String.concat ","
			   (List.map
			      (fun (f,l) ->
				 encaps(f ^ ","
					^ encaps(String.concat "," l))
			      ) (get_fields_flags prog)))
		      )
		   )
      :: !gen_insts;
      gen_insts := (gen_inst_set "methods_flags_per_classes"
		      (encaps(String.concat ","
				(List.map
				   (fun x ->
				      match x with
					| [] -> "()"
					| _ ->
					    encaps(
					      String.concat ","
						(List.map
						   (fun (m,l) ->
						      encaps(m ^ ","
							     ^ encaps(String.concat "," l))
						   ) x)
					    )
				   )
				   (get_methods_flags_per_classes prog)))))
      :: !gen_insts
    );List.rev !gen_insts
    
let import arg =
  match arg with
    | "java_bytecode" -> import_java_bytecode ()
    | _ -> failwith (
	"Import '" ^ arg ^ "' does not make part of the language")
