(** Provides a syntax tree representation for the language. *)

type insts =
  | Inst of inst * insts * string * int
  | Nop
and inst =
  | Import of string
  | Load of string
  | Genrule of expr * expr
(**********************************************)
  | GenruleCond of expr * expr * expr
(**********************************************)
  | Set of string * expr
  | Setn of string * expr * expr
  | Defun of string * string list * insts
  | LetIn of (string * expr) list * insts
  | If of expr * insts * insts
  | For of string * range * insts
  | Expr of expr
and expr = simple_expr * string * int
and simple_expr =
  | BaseExpr of base_expr
  | BaseTerm of string option * expr list
  | BaseConstructor of string * bool
  | BaseVar of string
  | Eval of string
  | Seq of string * bool * expr
  | CSeq of string * expr
  | USeq of string * expr
  | Getn of expr * expr
  | Subsn of expr * expr * expr
  | Add of expr list
  | Op of expr * expr
  | Mem of expr * expr
  | Cons of expr * expr
  | Len of expr
  | Depth of expr
  | Not of expr
  | And of expr list
  | Or of expr list
  | Concat of string * expr
  | FunCall of string * expr list
  | Equal of expr * expr
  | Sup of expr * expr
  | Inf of expr * expr
  | Different of expr * expr
(*********************************)
  | Plus of expr * expr
  | Minus of expr * expr
  | Times of expr * expr 
(*********************************)
  | Ne of expr * expr
  | Eq of expr * expr
  | Ge of expr * expr 
  | Le of expr * expr
  | Gt of expr * expr
  | Lt of expr * expr
(*********************************)
  | Et of expr * expr

and base_expr =
  | Term of string option * base_expr list
  | Constructor of string
  | Var of string
and range =
  | Range of expr
  | DblRange of expr * expr


val unit : base_expr

val string_of_base_expr : base_expr -> string

val int_of_base_expr : base_expr -> int

val keystring_of_expr : simple_expr -> string

val keystring_of_inst : inst -> string

val term_getn : base_expr -> int -> base_expr

val term_setn : base_expr -> int -> base_expr -> base_expr

val term_add : base_expr -> base_expr -> base_expr

val term_symboldepth : string -> base_expr -> int

val term_and : base_expr -> base_expr -> base_expr

val term_or : base_expr -> base_expr -> base_expr
