let rec term_of_pos_int (o:string) (c:string) (i : int) =
  if i = 0
  then (o,c)
  else
    let (o,c) = term_of_pos_int "" "" (i lsr 1)
    in
      if i land 1 = 1
      then (o^o^"succ(",")"^c^c)
      else (o^o,c^c)
let rec term_of_neg_int (o:string) (c:string) (i : int) =
  if i = 0
  then (o,c)
  else
    let (o,c) = term_of_neg_int "" "" (i / 2)
    in
      if i land 1 = 1
      then (o^o^"pred(",")"^c^c)
      else (o^o,c^c)

let term_of_int (i : int) =
  if i >= 0
  then
    let (o,c)=term_of_pos_int "" "" i
    in o^"zero"^c
  else
    let (o,c)=term_of_neg_int "" "" i
    in o^"zero"^c

let term_of_int32 (i : int32) =
  if i >= 0l
  then
    let (o,c) = term_of_pos_int "" "" (Int32.to_int (Int32.shift_right i 1))
    in
      if Int32.logand i 1l = 1l
      then o^o^"succ(zero)"^c^c
      else o^o^"zero"^c^c
  else
    let (o,c) = term_of_neg_int "" "" (Int32.to_int (Int32.div i 2l))
    in
      if Int32.logand i 1l = 1l
      then o^o^"pred(zero)"^c^c
      else o^o^"zero"^c^c

let term_of_int64 (i : int64) =
  let (lsr) = Int64.shift_right_logical
  and (land) = Int64.logand
  and (/) = Int64.div
  in let rec term_of_pos_int64 (o:string) (c:string) (i : int64) =
    if i = 0L
    then (o,c)
    else
      let (o,c) = term_of_pos_int64 "" "" (i lsr 1)
      in
	if i land 1L = 1L
	then (o^o^"succ(",")"^c^c)
	else (o^o,c^c)
  in let rec term_of_neg_int64 (o:string) (c:string) (i : int64) =
    if i = 0L
    then (o,c)
    else
      let (o,c) = term_of_neg_int64 "" "" (i / 2L)
      in
	if i land 1L = 1L
	then (o^o^"pred(",")"^c^c)
	else (o^o,c^c)
  in let (o,c)=if i>=0L then term_of_pos_int64 "" "" i else term_of_neg_int64 "" "" i
  in o^"zero"^c
